import java.io.*;

import java.util.Iterator;
import java.util.Stack;
import java.util.EmptyStackException;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.ArrayList;

import com.jclark.xml.parse.*;
import com.jclark.xml.parse.io.*;
import com.jclark.xml.parse.EntityManagerImpl;


class Statsxp {
    static class Elinfo {
	String name;
	int count;
	int minlev;
	int charcount;
	int seen;
	boolean empty;
	TreeMap ptab;
	TreeMap ktab;
	TreeMap atab;

	Elinfo(String nm) {
	    name = nm;
	    count = minlev = charcount = 0;
	    seen = 0;
	    empty = true;
	    ptab = new TreeMap();
	    ktab = new TreeMap();
	    atab = new TreeMap();
	}
    }

    static class Int {
	public int val;

	Int (int i) {
	    val = i;
	}
    }

    static HashMap eltab;
    static Stack elstack;
    static int global_seen = 0;
    static String root;

    static class Stats extends ApplicationImpl {
	public void startElement(StartElementEvent event)
	    throws IOException
	{
	    String name = event.getName();

	    Elinfo inf = (Elinfo) eltab.get(name);
	    if (inf == null) {
		inf = new Elinfo(name);
		inf.seen = global_seen++;
		eltab.put(name, inf);
	    }

	    inf.count++;

	    try {
		String parent = (String) elstack.peek();
		Int pcnt = (Int) inf.ptab.get(parent);
		if (pcnt == null) {
		    inf.ptab.put(parent, new Int(1));
		}
		else {
		    pcnt.val++;
		}

		Elinfo pinf = (Elinfo) eltab.get(parent);
		Int kcnt = (Int) pinf.ktab.get(name);
		if (kcnt == null) {
		    pinf.ktab.put(name, new Int(1));
		}
		else {
		    kcnt.val++;
		}
		pinf.empty = false;
	    }
	    catch (EmptyStackException e) {
		root = name;
	    }
	    
	    // Attribute handling

	    int n = event.getAttributeCount();
	    int i;
	    for (i = 0; i < n; i++) {
		String attnm = event.getAttributeName(i);

		Int acnt = (Int) inf.atab.get(attnm);
		if (acnt == null) {
		    inf.atab.put(attnm, new Int(1));
		}
		else {
		    acnt.val++;
		}
	    }

	    elstack.push(name);
	}

	public void endElement(EndElementEvent event)
	    throws IOException, EmptyStackException
	{
	    elstack.pop();
	}

	public void characterData(CharacterDataEvent event)
	    throws IOException
	{
	    String elname = (String) elstack.peek();
	    Elinfo inf = (Elinfo) eltab.get(elname);
	    inf.empty = false;
	    inf.charcount += event.getLength();
	}
    }

    static class ElComp implements java.util.Comparator {
	public int compare(Object ao, Object bo) {
	    Elinfo a = (Elinfo) ao;
	    Elinfo b = (Elinfo) bo;
	    int c1 = a.minlev - b.minlev;
	    if (c1 != 0)
		return c1;
	    return a.seen - b.seen;
	}

	public boolean equals(Object ao, Object bo) {
	    Elinfo a = (Elinfo) ao;
	    Elinfo b = (Elinfo) bo;
	    return ((a.minlev == b.minlev) && (a.seen == b.seen));
	}
    }

    static void set_minlev(String name, int level) {
	Elinfo inf = (Elinfo) eltab.get(name);

	if (inf.minlev == 0 || inf.minlev > level) {
	    int newlev = level + 1;
	    inf.minlev = level;
	    Iterator iter = inf.ktab.keySet().iterator();
	    while (iter.hasNext()) {
		String kname = (String) iter.next();
		set_minlev(kname, newlev);
	    }
	}
    }

    static void showtab(String label, TreeMap tab, boolean dosum) {
	if (tab.size() == 0)
	    return;
	System.out.println("\n   " + label + ":");
	int sum = 0;

	Iterator step = tab.keySet().iterator();
	while (step.hasNext()) {
	    String name = (String) step.next();
	    int cnt = ((Int) tab.get(name)).val;
	    sum += cnt;
	    System.out.print("      ");
	    System.out.print(name);
	    String cstr = Integer.toString(cnt);
	    int lim = 22 - name.length() + 5 - cstr.length();
	    int i;
	    for (i = 0; i < lim; i++)
		System.out.print(" ");
	    System.out.println(cstr);
	}

	if (dosum && tab.size() > 1) {
	    System.out.println("                            =====");
	    String sstr = Integer.toString(sum);
	    int lim = 33 - sstr.length();
	    int i;
	    for (i = 0; i < lim; i++)
		System.out.print(" ");
	    System.out.println(sstr);
	}
    }

    public static void main(String [] args)
    {
	if (args.length < 1) {
	    System.err.println("Missing file argument");
	    System.exit(1);
	}

	ParserImpl p = new ParserImpl();
	Stats app = new Stats();

	eltab = new HashMap(128);
	elstack = new Stack();

	p.setApplication(app);

	try {
	    OpenEntity doc = EntityManagerImpl.openFile(args[0]);
	    p.parseDocument(doc);
	}
	catch (FileNotFoundException e) {
	    System.err.println("Failed to open " + e.getMessage());
	    System.exit(1);
	}
	catch (NotWellFormedException e) {
	    System.err.println(" not well formed: " + e.getMessage());
	    System.exit(1);
	}
	catch (IOException e) {
	    e.printStackTrace();
	    System.exit(1);
	}

	set_minlev(root, 1);

	ArrayList sorted = new ArrayList(eltab.values());
	java.util.Collections.sort(sorted, new ElComp());
	Iterator step = sorted.iterator();

	while (step.hasNext()) {
	    Elinfo inf = (Elinfo) step.next();
	    System.out.println("\n================\n" + inf.name
			       + ": " + inf.count);
	    if (inf.charcount > 0)
		System.out.println("Had " + inf.charcount
				   + " bytes of character data");
	    if (inf.empty)
		System.out.println("Always empty");

	    showtab("Parents", inf.ptab, false);
	    showtab("Children", inf.ktab, true);
	    showtab("Attributes", inf.atab, false);
	}
    }
}
